;Spc2Midi - Display formats
;============================================================

section .bss
; Offscreen pixel buffer which the display routine has full reign to draw in
; The 32k buffer is then later copied to the screen.
alignb 4
SpcViewBuffer.SizeOf equ 128*256
SpcViewBuffer:       resb SpcViewBuffer.SizeOf

; A table for the display format to hold whatever various values it needs to
; The table is twice the width of the window (256*2)
SpcViewVars:
SpcViewVols             resb 128
SpcViewPreVols          resb 128
SpcViewColors           resb 128
SpcViewPreColors        resb 128
SpcViewPitches          resb 8
SpcViewVars.SizeOf      equ $-SpcViewVars

; Bit table to keep track of values in the above have changed
SpcViewChanges.SizeOf   equ (128*256)/8
SpcViewChanges          resb SpcViewChanges.SizeOf

;
section .data
align 4
SpcView:
.Handler:       dd SpcViewDrawers.Hns
;.Handler:       dd SpcViewDrawers.Vb
.ScreenCopier:  dd SpcViewCopy.PartialV;Complete;
.Row:
.Col:           dd 0    ;row or column

section .text
;
SpcViewSetMode:
    ;select case
.ClearTables:
    mov edi,SpcViewBuffer
    mov ecx,(SpcViewBuffer.SizeOf+SpcViewVars.SizeOf)/4
.Clear:
    cld
    xor eax,eax
    rep stosd
    mov ecx,SpcViewChanges.SizeOf/4
  %if SpcViewBuffer+SpcViewBuffer.SizeOf+SpcViewVars.SizeOf != SpcViewChanges
    %error SpcViewChanges must come after SpcViewVars
  %endif
    dec eax
    rep stosd
    ret
.ClearTable:
    mov edi,SpcViewVars
    mov ecx,SpcViewVars.SizeOf/4
    jmp short .Clear

;
SpcViewDrawers:
; Various display formats to view the notes in

;
; Horizontal Note Sheet
;
; Left to right.
.Hns:
    mov ebx,[SpcView.Col]
    mov ecx,64                  ;128/2
    mov edx,ebx
    inc edx                     ;row = (row + 1) and 255
    bts dword [SpcViewChanges],ebx
    and edx,255
    lea edi,[SpcViewBuffer+ebx]
    mov [SpcView.Col],edx
.HnsNextBar:
    mov byte [edi],0
    mov byte [edi+256],1
    add edi,512
    dec ecx
    jnz .HnsNextBar

;Draw dashed song position bar
    mov ecx,64                  ;128/2
    bts dword [SpcViewChanges],edx
    lea edi,[SpcViewBuffer+edx+256]
.HnsNextHightlight:
    mov byte [edi],4
    add edi,512
    dec ecx
    jnz .HnsNextHightlight

    mov ecx,7
    lea edi,[SpcViewBuffer+ebx]
.HnsNextVoice:
    ; if voice is on
    ;   if
    ;     voice was turned on, set decay to highest
    ;   else
    ;     decay--
    ;   endif
    ;   if voice not active (user turned it off)
    ;     show dim
    ;   else
    ;     show at current decay level
    ;   endif
    ; endif
    bt dword [MainSongPos.Active],ecx
    jnc .HnsSkipVoice
    bt dword [MutedVoices],ecx
    jnc .HnsSkipVoice

    mov edx,5
    bt dword [MainSongPos.Kon],ecx
    jc .HnsSetDecay
    mov dl,[SpcViewVols+ecx]    ;get note decay
    test dl,dl
    jle .HnsNoMoreDecay
    dec dl
.HnsSetDecay:
    mov [SpcViewVols+ecx],dl    ;set new note decay
.HnsNoMoreDecay:

    movzx ebx,byte [MainSongPos.Pitch+ecx] ;get note value
    xor ebx,127                 ;reverse vertically so that notes are not upside-down
    shl ebx,8                   ;*256, width of buffer
    mov al,[RainbowColorTable+ecx] ;color formula = { voice * 15 + 14 }
    add al,dl                   ;add voice's note decay to get rainbow shade
    add al,9
    mov [edi+ebx],al
.HnsSkipVoice:
    dec ecx
    jns .HnsNextVoice


%if 0
; display thin keyboard at bottom

; turn off old keys
    xor edx,edx
    mov edi,[MainSongPos.Kof]
    mov ecx,7                   ;start with last voice
    xor eax,eax
.HnskNextKof:
    mov dl,[SpcViewPitches+ecx] ;get previous pitch
    bt edi,ecx
    jc .HnskKof
    bt dword [MainSongPos.Active],ecx
    jnc .HnskSkipKof
    cmp [MainSongPos.Pitch+ecx],dl ;check if key changed
    je .HnskSkipKof
.HnskKof:
    shl edx,1
    mov [SpcViewBuffer+31744+edx],ax
    mov [SpcViewBuffer+32000+edx],ax
    mov [SpcViewBuffer+32256+edx],ax
    bts dword [SpcViewChanges],edx
.HnskSkipKof:
    dec ecx
    jns .HnskNextKof

; display new/changed keys
    ;xor edx,edx
    mov edi,[MainSongPos.Active]
    mov ecx,7                   ;start with last voice
    and edi,[MutedVoices]
.HnskNextKon:
    bt edi,ecx
    jnc .HnskSkipKon
    mov al,[SpcViewVols+ecx]    ;get note decay
    mov dl,[MainSongPos.Pitch+ecx]
    cmp [SpcViewPitches+ecx],dl ;check if key changed
    jne .HnskKonChange
    cmp [SpcViewPreVols+ecx],al ;compare with previous note decay
    je .HnskSkipKon
.HnskKonChange:
    mov [SpcViewPreVols+ecx],al ;set note decay
    add al,9
    mov [SpcViewPitches+ecx],dl
    add al,[RainbowColorTable+ecx] ;color formula = { voice * 15 + 14 }
    shl edx,1
    mov ah,al
    mov [SpcViewBuffer+31744+edx],ax
    mov [SpcViewBuffer+32000+edx],ax
    mov [SpcViewBuffer+32256+edx],ax
    bts dword [SpcViewChanges],edx

.HnskSkipKon:
    dec ecx
    jns .HnskNextKon
%endif

    ret

%if 0
;
; Vertical Note Sheet
;
; Top to bottom.
.Vns:
    ;call .NsReadBuffer
    cld
    mov ebx,[SpcView.Pos]
    bts dword [SpcViewChanges],ebx
    mov edx,ebx
    shl ebx,8                   ;*256
    inc edx                     ;row = (row + 1) and 127
    lea edi,[SpcViewBuffer+ebx]
    and edx,127
    mov ecx,256/4
    mov [SpcView.Pos],edx
    mov eax,00080008h           ;gray/black/gray/black
    rep stosd

    mov ecx,7
.VnsNextVoice:
    mov ebx,[SpcViewTable+ecx*4];get voice's note
    test ebx,ebx                ;test if null
    jz .VnsSkipVoice
    mov eax,ecx
    shl eax,1
    add eax,byte 32
    mov ah,al
    mov [edi-256 +ebx*2-2],ax
.VnsSkipVoice:
    dec ecx
    jns .VnsNextVoice
    ret
%endif

%if 0
;
; Instrument Sheet
.Is:
    mov ecx,7
    xor esi,esi
    xor edx,edx
    mov edi,[MainSongPos.Active]
    and edi,[MutedChannels]

; get previous pitch
; if voice off or (voice on & pitch change)
;   get previous sample number
;   set pixel dim
;   get note value
;   reverse value so that notes are not upside-down
;   write gray to (pitch * 256 + sample)
; endif
.IsKofNext:
    mov bh,[SpcViewTable+ecx]    ;get previous pitch
    bt dword [MainSongPos.Changed],ecx
    jnc .IsKofCheckPitch
    bt dword [MainSongPos.Active],ecx
    jc .IsKof
    bt edi,ecx
.IsKofCheckPitch:
    cmp [MainSongPos.Pitch+ecx],dl ;check if key changed
    je .HnskSkipKof
    jnc .IsNoChange
.IsKof:


    mov dl,[SpcViewTable+ecx+8]  ;get sample number
    not ebx                         ;reverse vertically
    and ebx,127<<8
    mov byte [SpcViewBuffer+ebx+edx],1
.IsOffEnd:



    mov dl,[SpcViewPitches+ecx] ;get previous pitch
    bt edi,ecx
    jc .HnskKof
    bt dword [MainSongPos.Active],ecx
    jnc .HnskSkipKof
    cmp [MainSongPos.Pitch+ecx],dl ;check if key changed
    je .HnskSkipKof


    ;     endif
    ;   flag column for redraw
    ; endloop while >= 0
    bts dword [SpcViewChanges],edx
.IsNoChange:
    add esi,byte 16             ;next voice in dsp regs
    dec ecx
    jns .IsNextVoice


    mov ecx,7
    xor esi,esi
.IsNextKon:
    ; if channel change
    bt edi,ecx
    jnc .IsNotOn

    bt dword [MainSongPos.Active],ecx
    jnc .IsNotOn
    ;     if note on
    ;       get sample number
    ;       get channel color, formula = channel * 15 + 14
    ;       reverse value so that notes are not upside-down
    ;       write color to (pitch * 256 + halfpitch \ 2)
    ;     else...
    mov bh,[MainSongPos.Pitch+ecx]
    mov dl,[MainSongPos.DspRegs+esi+DspReg.SourceNumber]  ;get sample number
    mov [SpcViewTable+ecx],bh
    mov [SpcViewTable+ecx+8],dl
    mov al,[RainbowColorTable+ecx]  ;color formula = voice * 15 + 14
    not ebx                         ;reverse vertically
    and ebx,127<<8
    add al,14
    mov [SpcViewBuffer+ebx+edx],al
    jmp short .IsOffEnd

.IsNoChange:
    add esi,byte 16             ;next voice in dsp regs
    dec ecx
    jns .IsNextVoice



    ret
%endif

%if 0
;
; Vertical Bars
.Vb:
; Decrement volume of all columns
    mov ecx,127                         ;start from last column
.VbNextDec:
    mov al,[SpcViewPreVols+ecx]         ;get previous column height
    and al,11111100b
    test al,al
    jz .VbBase
    sub al,4                            ;decay pixels
.VbBase:
    mov [SpcViewVols+ecx],al            ;set new height
    dec ecx
    jns .VbNextDec

; Add volumes of any active voices
    mov ecx,7                           ;start with last voice
    mov edi,[MainSongPos.Active]
    or edi,[MainSongPos.Kof]
    and edi,[MutedVoices]
.VbNextKon:
.VbKon:
    bt edi,ecx
    jnc .VbSkipVoice

    mov bl,[RainbowColorTable+ecx]      ;color formula = voice * 15 + 14
    movzx edx,byte [SpcViewPitches+ecx]
    add bl,9
    mov [SpcViewColors+edx],bl
    bt dword [MainSongPos.Kof],ecx
    jc .VbKof

    movzx edx,byte [MainSongPos.Pitch+ecx]
    add bl,5
    mov eax,[MainSongPos.Volume+ecx*4]
    shr eax,2                          ;/4
    bt dword [MainSongPos.Kon],ecx
    jnc .VbNoKon
    add eax,byte 4
.VbNoKon:
    cmp eax,127
    jbe .VbKonVolOk
    mov al,127
.VbKonVolOk:
    mov [SpcViewVols+edx],al
    mov [SpcViewPitches+ecx],dl
.VbKof:
    mov [SpcViewColors+edx],bl

.VbSkipVoice:
    dec ecx
    jns .VbNextKon

; Redraw columns
    xor ecx,ecx
    mov edx,127                         ;start from last column
    mov edi,254
.VbNextCol:

    ; if color changed, erase column
    mov al,[SpcViewColors+edx]          ;get color
    cmp [SpcViewPreColors+edx],al       ;compare with previous color
    je .VbColorSame
    mov [SpcViewPreColors+edx],al       ;set previous color
    ;bts dword [SpcViewChanges],edi     ;flag column as changed
    mov [SpcViewPreVols+edx],byte 0     ;zero previous column
    xor esi,esi                         ;black
    lea ebx,[SpcViewBuffer+edi]
    mov ecx,128
.VbNextClrRow:
    mov [ebx],si
    add ebx,256                         ;next row down
    dec ecx
    jg .VbNextClrRow
.VbColorSame:

    ; if column height changed, redraw difference
    mov cl,[SpcViewVols+edx]            ;get new column height
    mov bh,cl                           ;*256 (buffer width)
    sub cl,[SpcViewPreVols+edx]         ;get height difference
    mov [SpcViewPreVols+edx],bh         ;set previous to new
    je .VbSameVol

    ;mov al,[SpcViewColors+edx]         ;get voice color
    jnc .VbVolGreater
    sub bh,cl                           ;previous column - dif
    xor al,al                           ;erase column
    neg cl                              ;negative to positive
.VbVolGreater:
    not ebx                             ;invert vertical (top=bottom)
    and ebx,07F00h
    bts dword [SpcViewChanges],edi   ;flag column as changed
    lea ebx,[SpcViewBuffer+edi+ebx]
    mov ah,al                   ;make color word

.VbNextVolRow:
    mov [ebx],ax
    add ebx,256                         ;next row down
    dec ecx
    jg .VbNextVolRow
.VbSameVol:
    ;xor ecx,ecx

    sub edi,byte 2
    dec edx
    jns .VbNextCol

    ret

;
; Sample Pitch Bars
.Sb:
.SbNext:
    ret

%endif

;
; Temporary function until the GUI is completed.
SpcViewCopy:
.Complete:
    cld
    mov ebx,128
    mov esi,SpcViewBuffer
    mov edi,Screen.Buffer
.NextPixelRow:
    mov ecx,256/4
    rep movsd
    add edi,byte 64
    dec ebx
    jnz .NextPixelRow
    ret

.PartialV:
    cld
    push dword 128
    mov ebx,SpcViewBuffer
    mov edx,Screen.Buffer
.NextRow:
    xor ecx,ecx
.NextStrip:
    mov eax,[SpcViewChanges+ecx]
    mov esi,ebx
    mov edi,edx
.Next:
    test eax,15
    jz .SkipBlit
    movsd
    shr eax,4
    jnz .Next
    jmp short .AfterStripBlit
.SkipBlit:
    add esi,byte 4
    add edi,byte 4
    shr eax,4
    jnz .Next
.AfterStripBlit:
    add ebx,byte 32
    add edx,byte 32
    add ecx,byte 4
    cmp ecx,32
    jb .NextStrip
    add edx,byte 64
;    cmp edi,0B0000h
;    jb .ok
;    int3;
;.ok:
    dec dword [esp]
    jnz .NextRow
    pop ecx

;Reset change flags
    cld
    xor eax,eax
    ;mov ecx,SpcViewChanges.SizeOf/4
    mov ecx,256/4
    mov edi,SpcViewChanges
    rep stosd
    ret

;
; Temporary function until the GUI is completed.
; This does not belong in here.

ShowMutedVoices:
    mov edi,Screen.Buffer+(129*320)
    xor ecx,ecx
    mov ebx,10101010h
.NextChannel:
    bt dword [MutedVoices],ecx
    mov eax,ebx
    jnc .MutedChannel
    add eax,0E0E0E0Eh
.MutedChannel:
    mov [edi],eax
    mov [edi+320],eax
    shr eax,8
    mov [edi+4],eax
    mov [edi+324],eax

    inc ecx
    add edi,byte 8
    add ebx,1E1E1E1Eh
    cmp ecx,8
    jb .NextChannel
    ret
